/*
    Copyright (c) 2002-2010 Tampere University of Technology.

    This file is part of TTA-Based Codesign Environment (TCE).

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
 */
/**
 * @file QuartusProjectGenerator.hh
 *
 * Declaration of QuartusProjectGenerator class.
 *
 * @author Otto Esko 2010 (otto.esko-no.spam-tut.fi)
 * @note rating: red
 */

#include <string>
#include <vector>
#include <iostream>
#include <sstream>
#include <fstream>
#include "Exception.hh"
#include "FileSystem.hh"
#include "QuartusProjectGenerator.hh"
using std::string;
using std::ofstream;
using std::endl;

QuartusProjectGenerator::QuartusProjectGenerator(
    std::string toplevelEntity,
    PlatformIntegrator* integrator):
    ProjectFileGenerator(toplevelEntity,integrator) {
}

QuartusProjectGenerator::~QuartusProjectGenerator() {
}

void
QuartusProjectGenerator::writeProjectFiles() {
    
    writeQPFFile();

    writeQSFFile();

    writeScripts();
}


void
QuartusProjectGenerator::writeQPFFile() {
    
    string qpfFileName = toplevelEntity() + ".qpf";
    ofstream qpfFile;
    qpfFile.open(qpfFileName.c_str());
    if (!qpfFile) {
        string msg = "Couldn't open file " + qpfFileName + " for writing";
        IOException exc(__FILE__, __LINE__, "QuartusProjectGenerator", msg);
        throw exc;
    }
    qpfFile 
        << "# Generated by QuartusProject Generator" << endl << endl
        << "QUARTUS_VERSION = \"8.0\"" << endl << endl
        << "PROJECT_REVISION = \"" << toplevelEntity() << "\"" << endl;
    qpfFile.close();
}


void
QuartusProjectGenerator::writeQSFFile() {
    
    string qsfFileName = toplevelEntity() + ".qsf";
    ofstream qsfFile;
    qsfFile.open(qsfFileName.c_str());
    if (!qsfFile) {
        string msg = "Couldn't open file " + qsfFileName + " for writing";
        IOException exc(__FILE__, __LINE__, "QuartusProjectGenerator", msg);
        throw exc;
    }

    string deviceFamily = integrator()->deviceFamily();
    string deviceName = integrator()->deviceName();
    int fmax = integrator()->targetClockFrequency();
    qsfFile 
        << "# Generated by QuartusProjectGenerator" << endl << endl
        << "set_global_assignment -name FAMILY \"" << deviceFamily << "\""
        << endl
        << "set_global_assignment -name DEVICE " << deviceName << endl
        << "set_global_assignment -name TOP_LEVEL_ENTITY " << toplevelEntity()
        << endl
        << "set_global_assignment -name ORIGINAL_QUARTUS_VERSION 8.0" << endl
        << "set_global_assignment -name LAST_QUARTUS_VERSION 8.0" << endl
        << "set_global_assignment -name USE_GENERATED_PHYSICAL_CONSTRAINTS "
        << "OFF -section_id eda_palace" << endl
        << "set_global_assignment -name USE_CONFIGURATION_DEVICE ON" << endl
        << "set_global_assignment -name RESERVE_ALL_UNUSED_PINS \"AS "
        << "INPUT TRI-STATED\"" << endl
        << "set_global_assignment -name FMAX_REQUIREMENT \"" << fmax
        << " MHz\"" << endl;

    for (unsigned int i = 0; i < hdlFileList().size(); i++) {
        qsfFile << "set_global_assignment -name VHDL_FILE \"" 
                << hdlFileList().at(i) << "\"" << endl;
    }
    qsfFile << pinMappings() << endl;
    qsfFile.close();
}


void
QuartusProjectGenerator::writeScripts() {

    string synthesisScript = "quartus_synthesize.sh";
    ofstream synthesis;
    synthesis.open(synthesisScript.c_str());
    if (!synthesis) {
        string msg = "Couldn't open file " + synthesisScript + " for writing";
        IOException exc(__FILE__, __LINE__, "QuartusProjectGenerator", msg);
        throw exc;
    }
    synthesis
        << "#!/bin/bash" << endl
        << "# Generated by QuartusProjectGenerator" << endl << endl
        << "quartus_sh --flow compile " << toplevelEntity() << endl;
    synthesis.close();
    FileSystem::setFileExecutable(synthesisScript);

    string programmingScript = "quartus_program_fgpa.sh";
    ofstream programming;
    programming.open(programmingScript.c_str());
    if (!programming) {
        string msg = "Couldn't open file " + programmingScript
            + " for writing";
        IOException exc(__FILE__, __LINE__, "QuartusProjectGenerator", msg);
        throw exc;
    }
    programming
        << "#!/bin/bash" << endl
        << "# Generated by QuartusProjectGenerator" << endl << endl
        << "quartus_pgm -c USB-Blaster -m JTAG -o p\\;"
        << toplevelEntity() << ".sof" << endl << endl;
    programming.close();
    FileSystem::setFileExecutable(programmingScript);
}
